@@ -19,13 +19,14 @@ module Options.Applicative.Builder (
   -- creates a parser for an option called \"output\".
   subparser,
   argument,
-  argument',
   arguments,
   arguments1,
   flag,
   flag',
   switch,
   nullOption,
+  abortOption,
+  infoOption,
   strOption,
   option,
 
@@ -38,6 +39,7 @@ module Options.Applicative.Builder (
   showDefault,
   metavar,
   reader,
+  eitherReader,
   noArgError,
   ParseError(..),
   hidden,
@@ -59,6 +61,8 @@ module Options.Applicative.Builder (
   auto,
   str,
   disabled,
+  readerAbort,
+  readerError,
 
   -- * Builder for 'ParserInfo'
   InfoMod,
@@ -76,10 +80,19 @@ module Options.Applicative.Builder (
   disambiguate,
   showHelpOnError,
   noBacktrack,
-  prefs
+  noIntersperse,
+  prefs,
+
+  -- * Types
+  Mod,
+  ReadM,
+  OptionFields,
+  FlagFields,
+  ArgumentFields,
+  CommandFields
   ) where
 
-import Control.Applicative (pure, (<|>))
+import Control.Applicative (pure, (<|>), many, some)
 import Data.Monoid (Monoid (..)
 #if __GLASGOW_HASKELL__ > 702
   , (<>)
@@ -87,7 +100,6 @@ import Data.Monoid (Monoid (..)
   )
 
 import Options.Applicative.Builder.Completer
-import Options.Applicative.Builder.Arguments
 import Options.Applicative.Builder.Internal
 import Options.Applicative.Common
 import Options.Applicative.Types
@@ -98,7 +110,7 @@ import Options.Applicative.Types
 auto :: Monad m => Read a => String -> m a
 auto arg = case reads arg of
   [(r, "")] -> return r
-  _         -> fail "Cannot parse value"
+  _         -> fail $ "cannot parse value `" ++ arg ++ "'"
 
 -- | String 'Option' reader.
 str :: Monad m => String -> m String
@@ -106,7 +118,7 @@ str = return
 
 -- | Null 'Option' reader. All arguments will fail validation.
 disabled :: Monad m => String -> m a
-disabled = const . fail $ "Disabled option"
+disabled = const . fail $ "disabled option"
 
 -- modifiers --
 
@@ -135,9 +147,13 @@ help :: String -> Mod f a
 help s = optionMod $ \p -> p { propHelp = s }
 
 -- | Specify the 'Option' reader.
-reader :: (String -> Either ParseError a) -> Mod OptionFields a
+reader :: (String -> ReadM a) -> Mod OptionFields a
 reader f = fieldMod $ \p -> p { optReader = f }
 
+-- | Specify the 'Option' reader as a function in the 'Either' monad.
+eitherReader :: (String -> Either String a) -> Mod OptionFields a
+eitherReader f = reader (either readerError return . f)
+
 -- | Specify the error to display when no argument is provided to this option.
 noArgError :: ParseError -> Mod OptionFields a
 noArgError e = fieldMod $ \p -> p { optNoArgError = e }
@@ -181,9 +197,25 @@ completer f = fieldMod $ modCompleter (`mappend` f)
 subparser :: Mod CommandFields a -> Parser a
 subparser m = mkParser d g rdr
   where
-    Mod f d g = m `mappend` metavar "COMMAND"
-    CommandFields cmds = f (CommandFields [])
-    rdr = CmdReader (map fst cmds) (`lookup` cmds)
+    Mod _ d g = m `mappend` metavar "COMMAND"
+    rdr = uncurry CmdReader (mkCommand m)
+
+-- | Builder for an argument parser.
+argument :: (String -> Maybe a) -> Mod ArgumentFields a -> Parser a
+argument p (Mod f d g) = mkParser d g (ArgReader rdr)
+  where
+    ArgumentFields compl = f (ArgumentFields mempty)
+    rdr = CReader compl p
+
+-- | Builder for an argument list parser. All arguments are collected and
+-- returned as a list.
+arguments :: (String -> Maybe a) -> Mod ArgumentFields a -> Parser [a]
+arguments r m = many (argument r m)
+
+-- | Like `arguments`, but require at least one argument.
+arguments1 :: (String -> Maybe a) -> Mod ArgumentFields a -> Parser [a]
+arguments1 r m = some (argument r m)
+
 -- | Builder for a flag parser.
 --
 -- A flag that switches from a \"default value\" to an \"active value\" when
@@ -229,6 +261,23 @@ nullOption m = mkParser d g rdr
     crdr = CReader (optCompleter fields) (optReader fields)
     rdr = OptReader (optNames fields) crdr (optNoArgError fields)
 
+-- | An option that always fails.
+--
+-- When this option is encountered, the option parser immediately aborts with
+-- the given parse error.  If you simply want to output a message, use
+-- 'infoOption' instead.
+abortOption :: ParseError -> Mod OptionFields (a -> a) -> Parser (a -> a)
+abortOption err m = nullOption . (`mappend` m) $ mconcat
+  [ reader (const (ReadM (Left err)))
+  , noArgError err
+  , value id
+  , metavar ""
+  , hidden ]
+
+-- | An option that always fails and displays a message.
+infoOption :: String -> Mod OptionFields (a -> a) -> Parser (a -> a)
+infoOption = abortOption . InfoMsg
+
 -- | Builder for an option taking a 'String' argument.
 strOption :: Mod OptionFields String -> Parser String
 strOption m = nullOption $ reader str `mappend` m
@@ -300,6 +349,9 @@ showHelpOnError = PrefsMod $ \p -> p { prefShowHelpOnError = True }
 noBacktrack :: PrefsMod
 noBacktrack = PrefsMod $ \p -> p { prefBacktrack = False }
 
+noIntersperse :: PrefsMod
+noIntersperse = PrefsMod $ \p -> p { prefIntersperse = False }
+
 prefs :: PrefsMod -> ParserPrefs
 prefs m = applyPrefsMod m base
   where
@@ -307,7 +359,8 @@ prefs m = applyPrefsMod m base
       { prefMultiSuffix = ""
       , prefDisambiguate = False
       , prefShowHelpOnError = False
-      , prefBacktrack = True }
+      , prefBacktrack = True
+      , prefIntersperse = True }
 
 -- convenience shortcuts
 
