@@ -1,5 +1,7 @@
 module Options.Applicative.Builder.Completer
-  ( listIOCompleter
+  ( Completer
+  , mkCompleter
+  , listIOCompleter
   , listCompleter
   , bashCompleter
   ) where
@@ -20,8 +22,7 @@ listCompleter = listIOCompleter . pure
 
 bashCompleter :: String -> Completer
 bashCompleter action = Completer $ \word -> do
-  let cmd = unwords ["compgen", "-A", action, word]
-
+  let cmd = unwords ["compgen", "-A", action, "--", word]
   result <- tryIO $ readProcess "bash" ["-c", cmd] ""
   return . lines . either (const []) id $ result
 
