@@ -21,6 +21,7 @@ module Options.Applicative.Common (
   --
   Parser,
   liftOpt,
+  showOption,
 
   -- * Program descriptions
   --
@@ -44,15 +45,21 @@ module Options.Applicative.Common (
   optionNames
   ) where
 
-import Control.Applicative (pure, (<*>), (<$>), (<|>), empty)
-import Control.Monad (guard, msum)
+import Control.Applicative (pure, (<*>), (<$>), (<|>), (<$))
+import Control.Monad (guard, mzero, msum, when, liftM, MonadPlus)
+import Control.Monad.Trans.Class (lift)
+import Control.Monad.Trans.State (StateT(..), get, put, runStateT)
 import Data.List (isPrefixOf)
-import Data.Maybe (maybeToList, isJust)
+import Data.Maybe (maybeToList, isJust, isNothing)
 import Data.Monoid (Monoid(..))
 
 import Options.Applicative.Internal
 import Options.Applicative.Types
 
+showOption :: OptName -> String
+showOption (OptLong n) = "--" ++ n
+showOption (OptShort n) = '-' : [n]
+
 optionNames :: OptReader a -> [OptName]
 optionNames (OptReader names _ _) = names
 optionNames (FlagReader names _) = names
@@ -76,105 +83,141 @@ instance Monoid MatchResult where
   mappend m@(Match _) _ = m
   mappend _ m = m
 
-type Matcher m a = [String] -> m (a, [String])
+type Args = [String]
 
-optMatches :: MonadP m => Bool -> OptReader a -> String -> Maybe (Matcher m a)
-optMatches disambiguate opt arg = case opt of
-  OptReader names rdr no_arg_err -> do
-    (arg1, val) <- parsed
-    guard $ has_name arg1 names
-    return $ \args -> do
-      let mb_args = uncons $ maybeToList val ++ args
-      let missing_arg = missingArgP no_arg_err (crCompleter rdr)
-      (arg', args') <- maybe missing_arg return mb_args
-      r <- liftEither (crReader rdr arg')
-      return (r, args')
-  FlagReader names x -> do
-    (arg1, Nothing) <- parsed
-    guard $ has_name arg1 names
-    return $ \args -> return (x, args)
+argMatches :: MonadP m => OptReader a -> String
+           -> Maybe (StateT Args m a)
+argMatches opt arg = case opt of
   ArgReader rdr -> do
     result <- crReader rdr arg
-    return $ \args -> return (result, args)
+    Just $ return result
   CmdReader _ f ->
-    flip fmap (f arg) $ \subp args -> do
+    flip fmap (f arg) $ \subp -> StateT $ \args -> do
       setContext (Just arg) subp
       prefs <- getPrefs
       let runSubparser
-            | prefBacktrack prefs = runParser
+            | prefBacktrack prefs = \p a -> do
+                policy <- getPolicy
+                runParser policy p a
             | otherwise = \p a
             -> (,) <$> runParserFully p a <*> pure []
       runSubparser (infoParser subp) args
+  _ -> Nothing
+
+optMatches :: MonadP m => Bool -> OptReader a -> OptWord -> Maybe (StateT Args m a)
+optMatches disambiguate opt (OptWord arg1 val) = case opt of
+  OptReader names rdr no_arg_err -> do
+    guard $ has_name arg1 names
+    Just $ do
+      args <- get
+      let mb_args = uncons $ maybeToList val ++ args
+      let missing_arg = lift $ missingArgP no_arg_err (crCompleter rdr)
+      (arg', args') <- maybe missing_arg return mb_args
+      put args'
+      case runReadM (crReader rdr arg') of
+        Left e -> lift $ errorFor arg1 e
+        Right r -> return r
+  FlagReader names x -> do
+    guard $ has_name arg1 names
+    guard $ isNothing val
+    Just $ return x
+  _ -> Nothing
   where
-    parsed =
-      case arg of
-        '-' : '-' : arg1 ->
-          Just $
-          case span (/= '=') arg1 of
-            (_, "") -> (OptLong arg1, Nothing)
-            (arg1', _ : rest) -> (OptLong arg1', Just rest)
-        '-' : arg1 ->
-          case arg1 of
-            [] -> Nothing
-            (a : rest) -> Just (OptShort a, if null rest then Nothing else Just rest)
-        _ -> Nothing
+    errorFor name (ErrorMsg msg) =
+      errorP (ErrorMsg ("option " ++ showOption name ++ ": " ++ msg))
+    errorFor _ e = errorP e
+
     has_name a
       | disambiguate = any (isOptionPrefix a)
       | otherwise = elem a
 
-stepParser :: MonadP m => ParserPrefs -> Parser a -> String -> [String] -> [m (Parser a, [String])]
-stepParser _ (NilP _) _ _ = []
-stepParser prefs (OptP opt) arg args =
-  case optMatches disambiguate (optMain opt) arg of
-    Just matcher -> pure $ do
-      (r, args') <- matcher args
-      return (pure r, args')
-    Nothing -> empty
-  where
-    disambiguate = prefDisambiguate prefs
-                && optVisibility opt > Internal
-stepParser prefs (MultP p1 p2) arg args = msum
-  [ flip map (stepParser prefs p1 arg args) $ \m ->
-      do (p1', args') <- m
-         return (p1' <*> p2, args')
-  , flip map (stepParser prefs p2 arg args) $ \m ->
-      do (p2', args') <- m
-         return (p1 <*> p2', args') ]
-stepParser prefs (AltP p1 p2) arg args = msum
-  [ stepParser prefs p1 arg args
-  , stepParser prefs p2 arg args ]
-stepParser prefs (BindP p k) arg args =
-  flip map (stepParser prefs p arg args) $ \m -> do
-    (p', args') <- m
-    x <- liftMaybe $ evalParser p'
-    return (k x, args')
+isArg :: OptReader a -> Bool
+isArg (ArgReader _) = True
+isArg _ = False
+
+data OptWord = OptWord OptName (Maybe String)
+
+parseWord :: String -> Maybe OptWord
+parseWord ('-' : '-' : w) = Just $ let
+  (opt, arg) = case span (/= '=') w of
+    (_, "") -> (w, Nothing)
+    (w', _ : rest) -> (w', Just rest)
+  in OptWord (OptLong opt) arg
+parseWord ('-' : w) = case w of
+  [] -> Nothing
+  (a : rest) -> Just $ let
+    arg = rest <$ guard (not (null rest))
+    in OptWord (OptShort a) arg
+parseWord _ = Nothing
+
+searchParser :: Monad m
+             => (forall r . Option r -> NondetT m r)
+             -> Parser a -> NondetT m (Parser a)
+searchParser _ (NilP _) = mzero
+searchParser f (OptP opt) = liftM pure (f opt)
+searchParser f (MultP p1 p2) = foldr1 (<!>)
+  [ do p1' <- searchParser f p1
+       return (p1' <*> p2)
+  , do p2' <- searchParser f p2
+       return (p1 <*> p2') ]
+searchParser f (AltP p1 p2) = msum
+  [ searchParser f p1
+  , searchParser f p2 ]
+searchParser f (BindP p k) = do
+  p' <- searchParser f p
+  x <- hoistMaybe (evalParser p')
+  return (k x)
+
+searchOpt :: MonadP m => ParserPrefs -> OptWord -> Parser a
+          -> NondetT (StateT Args m) (Parser a)
+searchOpt pprefs w = searchParser $ \opt -> do
+  let disambiguate = prefDisambiguate pprefs
+                  && optVisibility opt > Internal
+  case optMatches disambiguate (optMain opt) w of
+    Just matcher -> lift matcher
+    Nothing -> mzero
+
+searchArg :: MonadP m => String -> Parser a
+          -> NondetT (StateT Args m) (Parser a)
+searchArg arg = searchParser $ \opt -> do
+  when (isArg (optMain opt)) cut
+  case argMatches (optMain opt) arg of
+    Just matcher -> lift matcher
+    Nothing -> mzero
+
+data ArgPolicy
+  = SkipOpts
+  | AllowOpts
+  deriving Eq
+
+stepParser :: MonadP m => ParserPrefs -> ArgPolicy -> String
+           -> Parser a -> NondetT (StateT Args m) (Parser a)
+stepParser pprefs SkipOpts arg p = case parseWord arg of
+  Just w -> searchOpt pprefs w p
+  Nothing -> searchArg arg p
+stepParser pprefs AllowOpts arg p = msum
+  [ searchArg arg p
+  , do w <- hoistMaybe (parseWord arg)
+       searchOpt pprefs w p ]
 
 -- | Apply a 'Parser' to a command line, and return a result and leftover
 -- arguments.  This function returns an error if any parsing error occurs, or
 -- if any options are missing and don't have a default value.
-runParser :: MonadP m => Parser a -> [String] -> m (a, [String])
-runParser p args = case args of
+runParser :: MonadP m => ArgPolicy -> Parser a -> Args -> m (a, Args)
+runParser policy p args = case args of
   [] -> exitP p result
+  ("--" : argt) -> runParser AllowOpts p argt
   (arg : argt) -> do
     prefs <- getPrefs
-    x <- tryP $ do_step prefs arg argt
-    case x of
-      Left e -> case (result, e) of
-        (Just r, ErrorMsg _) -> return r
-        _ -> errorP e
-      Right (p', args') -> runParser p' args'
+    (mp', args') <- do_step prefs arg argt
+    case mp' of
+      Nothing -> hoistMaybe result <|> parseError arg
+      Just p' -> runParser policy p' args'
   where
     result = (,) <$> evalParser p <*> pure args
-    do_step prefs arg argt
-      | prefDisambiguate prefs
-      = case parses of
-          [m] -> m
-          _   -> empty
-      | otherwise
-      = case parses of
-          [] -> parseError arg
-          (m : _) -> m
-      where parses = stepParser prefs p arg argt
+    do_step prefs arg argt = (`runStateT` argt)
+                           . disamb (not (prefDisambiguate prefs))
+                           $ stepParser prefs policy arg p
 
 parseError :: MonadP m => String -> m a
 parseError arg = errorP . ErrorMsg $ msg
@@ -183,9 +226,17 @@ parseError arg = errorP . ErrorMsg $ msg
       ('-':_) -> "Invalid option `" ++ arg ++ "'"
       _       -> "Invalid argument `" ++ arg ++ "'"
 
-runParserFully :: MonadP m => Parser a -> [String] -> m a
+getPolicy :: MonadP m => m ArgPolicy
+getPolicy = do
+  prefs <- getPrefs
+  return $ if prefIntersperse prefs
+             then SkipOpts
+             else AllowOpts
+
+runParserFully :: MonadP m => Parser a -> Args -> m a
 runParserFully p args = do
-  (r, args') <- runParser p args
+  policy <- getPolicy
+  (r, args') <- runParser policy p args
   guard $ null args'
   return r
 
