@@ -4,38 +4,46 @@ module Options.Applicative.Extra (
   --
   -- | This module contains high-level functions to run parsers.
   helper,
+  hsubparser,
   execParser,
-  execParserPure,
+  execParserMaybe,
   customExecParser,
+  customExecParserMaybe,
+  execParserPure,
   usage,
   ParserFailure(..),
   ) where
 
-import Control.Applicative ((<$>), (<|>))
+import Control.Applicative ((<$>), (<|>), (<**>))
 import Data.Monoid (mconcat)
 import System.Environment (getArgs, getProgName)
 import System.Exit (exitWith, ExitCode(..))
 import System.IO (hPutStr, stderr)
 
 import Options.Applicative.BashCompletion
-import Options.Applicative.Common
 import Options.Applicative.Builder hiding (briefDesc)
+import Options.Applicative.Builder.Internal
+import Options.Applicative.Common
 import Options.Applicative.Help
 import Options.Applicative.Internal
-import Options.Applicative.Utils
 import Options.Applicative.Types
+import Options.Applicative.Utils
 
 -- | A hidden \"helper\" option which always fails.
 helper :: Parser (a -> a)
-helper = nullOption $ mconcat
-       [ long "help"
-       , reader (const (Left ShowHelpText))
-       , noArgError ShowHelpText
-       , short 'h'
-       , help "Show this help text"
-       , value id
-       , metavar ""
-       , hidden ]
+helper = abortOption ShowHelpText $ mconcat
+  [ long "help"
+  , short 'h'
+  , help "Show this help text" ]
+
+hsubparser :: Mod CommandFields a -> Parser a
+hsubparser m = mkParser d g rdr
+  where
+    Mod _ d g = m `mappend` metavar "COMMAND"
+    (cmds, subs) = mkCommand m
+    rdr = CmdReader cmds (fmap add_helper . subs)
+    add_helper pinfo = pinfo
+      { infoParser = infoParser pinfo <**> helper }
 
 -- | Run a program description.
 --
@@ -59,10 +67,28 @@ customExecParser pprefs pinfo = do
         _           -> hPutStr stderr msg
       exitWith c
 
+-- | Run a program description in pure code.
+--
+-- This function behaves like 'execParser', but can be called from pure code.
+-- Note that, in case of errors, no message is displayed, and this function
+-- simply returns 'Nothing'.
+--
+-- If you need to keep track of error messages, use 'execParserPure' instead.
+execParserMaybe :: ParserInfo a -> [String] -> Maybe a
+execParserMaybe = customExecParserMaybe (prefs idm)
+
+-- | Run a program description with custom preferences in pure code.
+--
+-- See 'execParserMaybe' for details.
+customExecParserMaybe :: ParserPrefs -> ParserInfo a -> [String] -> Maybe a
+customExecParserMaybe pprefs pinfo
+  = either (const Nothing) Just
+  . execParserPure pprefs pinfo
+
 data Result a = Result a
               | Extra ParserFailure
 
--- | A pure version 'execParser'.
+-- | The most general way to run a program description in pure code.
 execParserPure :: ParserPrefs       -- ^ Global preferences for this parser
                -> ParserInfo a      -- ^ Description of the program to run
                -> [String]          -- ^ Program arguments
@@ -90,12 +116,14 @@ parserFailure pprefs pinfo msg ctx = ParserFailure
            . show_help
            . add_error
            . add_usage names progn
-  , errExitCode = ExitFailure (infoFailureCode pinfo) }
+  , errExitCode = exit_code }
   where
-    add_usage names progn i = i
-      { infoHeader = vcat
-          ( header_line i ++
-            [ usage pprefs (infoParser i) ename ] ) }
+    add_usage names progn i = case msg of
+      InfoMsg _ -> i
+      _         -> i
+        { infoHeader = vcat
+            ( header_line i ++
+              [ usage pprefs (infoParser i) ename ] ) }
       where
         ename = unwords (progn : names)
     add_error i = i
@@ -103,6 +131,10 @@ parserFailure pprefs pinfo msg ctx = ParserFailure
     error_msg = case msg of
       ShowHelpText -> []
       ErrorMsg m   -> [m]
+      InfoMsg  m   -> [m]
+    exit_code = case msg of
+      InfoMsg _ -> ExitSuccess
+      _         -> ExitFailure (infoFailureCode pinfo)
     show_full_help = case msg of
       ShowHelpText -> True
       _            -> prefShowHelpOnError pprefs
