@@ -9,10 +9,14 @@ module Options.Applicative.Types (
   OptReader(..),
   OptProperties(..),
   OptVisibility(..),
+  ReadM(..),
+  readerAbort,
+  readerError,
   CReader(..),
   Parser(..),
   ParserM(..),
   Completer(..),
+  mkCompleter,
   ParserFailure(..),
   OptHelpInfo(..),
   OptTree(..),
@@ -30,13 +34,14 @@ module Options.Applicative.Types (
 
 import Control.Applicative
   (Applicative(..), Alternative(..), (<$>), optional)
-import Control.Monad (ap, liftM)
+import Control.Monad (ap, liftM, MonadPlus, mzero, mplus)
 import Control.Monad.Trans.Error (Error(..))
 import Data.Monoid (Monoid(..))
 import System.Exit (ExitCode(..))
 
 data ParseError
   = ErrorMsg String
+  | InfoMsg String
   | ShowHelpText
   deriving Show
 
@@ -59,9 +64,14 @@ instance Functor ParserInfo where
 -- | Global preferences for a top-level 'Parser'.
 data ParserPrefs = ParserPrefs
   { prefMultiSuffix :: String    -- ^ metavar suffix for multiple options
-  , prefDisambiguate :: Bool     -- ^ automatically disambiguate abbreviations (default: False)
-  , prefShowHelpOnError :: Bool  -- ^ always show help text on parse errors (default: False)
-  , prefBacktrack :: Bool        -- ^ backtrack to parent parser when a subcommand fails (default: True)
+  , prefDisambiguate :: Bool     -- ^ automatically disambiguate abbreviations
+                                 -- (default: False)
+  , prefShowHelpOnError :: Bool  -- ^ always show help text on parse errors
+                                 -- (default: False)
+  , prefBacktrack :: Bool        -- ^ backtrack to parent parser when a
+                                 -- subcommand fails (default: True)
+  , prefIntersperse :: Bool      -- ^ allow regular options and flags to occur
+                                 -- after arguments (default: True)
   }
 
 data OptName = OptShort !Char
@@ -99,7 +109,37 @@ data CReader m a = CReader
 instance Functor m => Functor (CReader m) where
   fmap f (CReader c r) = CReader c (fmap f . r)
 
-type OptCReader = CReader (Either ParseError)
+-- | A newtype over the 'Either' monad used by option readers.
+newtype ReadM a = ReadM
+  { runReadM :: Either ParseError a }
+
+instance Functor ReadM where
+  fmap f (ReadM m) = ReadM (fmap f m)
+
+instance Applicative ReadM where
+  pure = ReadM . Right
+  ReadM b <*> ReadM a = ReadM (b <*> a)
+
+instance Monad ReadM where
+  return = ReadM . Right
+  ReadM m >>= f = ReadM $ m >>= runReadM . f
+  fail = ReadM . Left . ErrorMsg
+
+instance MonadPlus ReadM where
+  mzero = ReadM $ Left (strMsg "")
+  mplus m1 m2 = case runReadM m1 of
+    Left _ -> m2
+    Right r -> return r
+
+-- | Abort option reader by exiting with a 'ParseError'.
+readerAbort :: ParseError -> ReadM a
+readerAbort = ReadM . Left
+
+-- | Abort option reader by exiting with an error message.
+readerError :: String -> ReadM a
+readerError = readerAbort . ErrorMsg
+
+type OptCReader = CReader ReadM
 type ArgCReader = CReader Maybe
 
 -- | An 'OptReader' defines whether an option matches an command line argument.
@@ -173,6 +213,9 @@ instance Alternative Parser where
 newtype Completer = Completer
   { runCompleter :: String -> IO [String] }
 
+mkCompleter :: (String -> IO [String]) -> Completer
+mkCompleter = Completer
+
 instance Monoid Completer where
   mempty = Completer $ \_ -> return []
   mappend (Completer c1) (Completer c2) =
