@@ -1,9 +1,15 @@
-{-# LANGUAGE Arrows #-}
+{-# LANGUAGE Arrows, CPP #-}
 module Examples.Cabal where
 
 import Options.Applicative
 import Options.Applicative.Arrows
 
+#if __GLASGOW_HASKELL__ <= 702
+import Data.Monoid
+(<>) :: Monoid a => a -> a -> a
+(<>) = mappend
+#endif
+
 data Args = Args CommonOpts Command
   deriving Show
 
@@ -32,10 +38,15 @@ data BuildOpts = BuildOpts
   { buildDir :: FilePath }
   deriving Show
 
+version :: Parser (a -> a)
+version = infoOption "0.0.0"
+  (  long "version"
+  <> help "Print version information" )
+
 parser :: Parser Args
 parser = runA $ proc () -> do
   opts <- asA commonOpts -< ()
-  cmds <- (asA . subparser)
+  cmds <- (asA . hsubparser)
             ( command "install"
               (info installParser
                     (progDesc "Installs a list of packages"))
@@ -48,7 +59,7 @@ parser = runA $ proc () -> do
            <> command "build"
               (info buildParser
                     (progDesc "Make this package ready for installation")) ) -< ()
-  A helper -< Args opts cmds
+  A version >>> A helper -< Args opts cmds
 
 commonOpts :: Parser CommonOpts
 commonOpts = CommonOpts
@@ -63,7 +74,7 @@ installParser :: Parser Command
 installParser = runA $ proc () -> do
   config <- asA configureOpts -< ()
   inst <- asA installOpts -< ()
-  A helper -< Install config inst
+  returnA -< Install config inst
 
 installOpts :: Parser InstallOpts
 installOpts = runA $ proc () -> do
@@ -74,13 +85,12 @@ installOpts = runA $ proc () -> do
              , instForce = force }
 
 updateParser :: Parser Command
-updateParser = runA $ proc () ->
-  A helper -< Update
+updateParser = pure Update
 
 configureParser :: Parser Command
 configureParser = runA $ proc () -> do
   config <- asA configureOpts -< ()
-  A helper -< Configure config
+  returnA -< Configure config
 
 configureOpts :: Parser ConfigureOpts
 configureOpts = runA $ proc () -> do
@@ -97,7 +107,7 @@ configureOpts = runA $ proc () -> do
 buildParser :: Parser Command
 buildParser = runA $ proc () -> do
   opts <- asA buildOpts -< ()
-  A helper -< Build opts
+  returnA -< Build opts
 
 buildOpts :: Parser BuildOpts
 buildOpts = runA $ proc () -> do
