@@ -1,8 +1,15 @@
+{-# LANGUAGE CPP #-}
 module Examples.Commands where
 
 import Data.List
 import Options.Applicative
 
+#if __GLASGOW_HASKELL__ <= 702
+import Data.Monoid
+(<>) :: Monoid a => a -> a -> a
+(<>) = mappend
+#endif
+
 data Sample
   = Hello [String]
   | Goodbye
