@@ -1,4 +1,4 @@
-{-# LANGUAGE TemplateHaskell #-}
+{-# LANGUAGE TemplateHaskell, CPP #-}
 module Main where
 
 import qualified Examples.Hello as Hello
@@ -6,12 +6,20 @@ import qualified Examples.Commands as Commands
 import qualified Examples.Cabal as Cabal
 import qualified Examples.Alternatives as Alternatives
 
+import Control.Monad
+import Data.List
 import Options.Applicative
 import System.Exit
 import Test.HUnit
 import Test.Framework.Providers.HUnit
 import Test.Framework.TH.Prime
 
+#if __GLASGOW_HASKELL__ <= 702
+import Data.Monoid
+(<>) :: Monoid a => a -> a -> a
+(<>) = mappend
+#endif
+
 run :: ParserInfo a -> [String] -> Either ParserFailure a
 run = execParserPure (prefs idm)
 
@@ -20,6 +28,13 @@ assertLeft x f = either f err x
   where
     err b = assertFailure $ "expected Left, got " ++ show b
 
+assertRight :: Either ParserFailure b -> (b -> Assertion) -> Assertion
+assertRight x f = either err f x
+  where
+    err (ParserFailure e _) = do
+      msg <- e "test"
+      assertFailure $ "unexpected parse error\n" ++ msg
+
 assertHasLine :: String -> String -> Assertion
 assertHasLine l s
   | l `elem` lines s = return ()
@@ -40,8 +55,8 @@ case_hello = checkHelpText "hello" Hello.opts ["--help"]
 case_modes :: Assertion
 case_modes = checkHelpText "commands" Commands.opts ["--help"]
 
-case_cabal :: Assertion
-case_cabal = checkHelpText "cabal" Cabal.pinfo ["configure", "--help"]
+case_cabal_conf :: Assertion
+case_cabal_conf = checkHelpText "cabal" Cabal.pinfo ["configure", "--help"]
 
 case_args :: Assertion
 case_args = do
@@ -219,6 +234,14 @@ case_arguments1_some = do
     Left _ -> assertFailure "unexpected parse error"
     Right r -> ["foo", "bar", "baz"] @=? r
 
+case_arguments_switch :: Assertion
+case_arguments_switch = do
+  let p =  switch (short 'x')
+        *> arguments str idm
+      i = info p idm
+      result = run i ["--", "-x"]
+  assertRight result $ \args -> ["-x"] @=? args
+
 case_issue_35 :: Assertion
 case_issue_35 = do
   let p =  flag' True (short 't' <> hidden)
@@ -242,5 +265,91 @@ case_backtracking = do
       result = execParserPure (prefs noBacktrack) i ["c", "-b"]
   assertLeft result $ \ _ -> return ()
 
+case_error_context :: Assertion
+case_error_context = do
+  let p = pk <$> option (long "port")
+             <*> option (long "key")
+      i = info p idm
+      result = run i ["--port", "foo", "--key", "291"]
+  case result of
+    Left (ParserFailure err _) -> do
+      msg <- err "test"
+      let errMsg = head $ lines msg
+      assertBool "no context in error message (option)"
+                 ("port" `isInfixOf` errMsg)
+      assertBool "no context in error message (value)"
+                 ("foo" `isInfixOf` errMsg)
+    Right val ->
+      assertFailure $ "unexpected result " ++ show val
+  where
+    pk :: Int -> Int -> (Int, Int)
+    pk = (,)
+
+condr :: MonadPlus m => (Int -> Bool) -> String -> m Int
+condr f arg = do
+  x <- auto arg
+  guard (f (x :: Int))
+  return x
+
+case_arg_order_1 :: Assertion
+case_arg_order_1 = do
+  let p = (,)
+          <$> argument (condr even) idm
+          <*> argument (condr odd) idm
+      i = info p idm
+      result = run i ["3", "6"]
+  assertLeft result $ \_ -> return ()
+
+case_arg_order_2 :: Assertion
+case_arg_order_2 = do
+  let p = (,,)
+        <$> argument (condr even) idm
+        <*> option (reader (condr even) <> short 'a')
+        <*> option (reader (condr odd) <> short 'b')
+      i = info p idm
+      result = run i ["2", "-b", "3", "-a", "6"]
+  case result of
+    Left _ -> assertFailure "unexpected parse error"
+    Right res -> (2, 6, 3) @=? res
+
+case_arg_order_3 :: Assertion
+case_arg_order_3 = do
+  let p = (,)
+          <$> (  argument (condr even) idm
+             <|> option (short 'n') )
+          <*> argument (condr odd) idm
+      i = info p idm
+      result = run i ["-n", "3", "5"]
+  case result of
+    Left _ ->
+      assertFailure "unexpected parse error"
+    Right res -> (3, 5) @=? res
+
+case_issue_47 :: Assertion
+case_issue_47 = do
+  let p = nullOption (long "test" <> reader r <> value 9) :: Parser Int
+      r _ = readerError "error message"
+      result = run (info p idm) ["--test", "x"]
+  assertLeft result $ \(ParserFailure err _) -> do
+    text <- head . lines <$> err "test"
+    assertBool "no error message"
+               ("error message" `isInfixOf` text)
+
+case_issue_50 :: Assertion
+case_issue_50 = do
+  let p = argument str (metavar "INPUT")
+          <* switch (long "version")
+      result = run (info p idm) ["--version", "test"]
+  assertRight result $ \r -> "test" @=? r
+
+case_intersperse_1 :: Assertion
+case_intersperse_1 = do
+  let p = arguments str (metavar "ARGS")
+          <* switch (short 'x')
+      result = execParserPure (prefs noIntersperse)
+                              (info p idm)
+                              ["a", "-x", "b"]
+  assertRight result $ \args -> ["a", "-x", "b"] @=? args
+
 main :: IO ()
 main = $(defaultMainGenerator)
